// Picassobot - Robotic Drawing Arm kit
// Copyright (c) 2014 - The Educated Robot

/**
 * This implementation of Canny Edge Detection is stripped down and
 * adapted from Tom Gibara's implementation at:
 * http://www.tomgibara.com/computer-vision/CannyEdgeDetector.java
**/

final static float GAUSSIAN_CUT_OFF = 0.005f;
final static float MAGNITUDE_SCALE = 100F;
final static float MAGNITUDE_LIMIT = 1000F;
final static int MAGNITUDE_MAX = (int) (MAGNITUDE_SCALE * MAGNITUDE_LIMIT);
final static float DEFAULT_LOWTHRESHOLD = 2.5; // 1.2 // Must be >= 0.
final static float DEFAULT_HIGHTHRESHOLD = 7.5; // 9.1 // Must be >= 0.
final static float DEFAULT_GAUSSIANKERNELRADIUS = 1.3; //2; // Must be >= 0.1. Need to * ratio of input width / output width.
final static int DEFAULT_GAUSSIANKERNELWIDTH = 16;//16; // Must be >= 2. Need to * ratio of input width / output width.

float lowThreshold = DEFAULT_LOWTHRESHOLD; // Must be >= 0.
float highThreshold = DEFAULT_HIGHTHRESHOLD; // Must be >= 0.
float gaussianKernelRadius = DEFAULT_GAUSSIANKERNELRADIUS;//2; // Must be >= 0.1. Need to * ratio of input width / output width.
int gaussianKernelWidth = DEFAULT_GAUSSIANKERNELWIDTH;//16; // Must be >= 2. Need to * ratio of input width / output width.
boolean contrastNormalized = false;

/* video
PImage edgeImg = null;
Capture webcam;

PImage img = null;
boolean processingImage = false;
  
int mFollowStackDepth = 100; // crash is at ~150 on Nexus One
  
int imageHeight;
int imageWidth;
int picsize;
int[] data;
int[] magnitude;
float[] xConv;
float[] yConv;
float[] xGradient;
float[] yGradient;


void updateProgress(String s) {
}

// NOTE: It is quite feasible to replace the implementation of this method
// with one which only loosely approximates the hypot function. I've tested
// simple approximations such as Math.abs(x) + Math.abs(y) and they work fine.
float hypot(float x, float y) {
    return (float) Math.hypot(x, y);
}
 
float gaussian(float x, float sigma) {
    return (float) Math.exp(-(x * x) / (2f * sigma * sigma));
  }
 
void follow(int x1, int y1, int i1, int threshold, int depth) {
    if( depth > mFollowStackDepth)  // don't run out of stack!
      return;
    
    int x0 = x1 == 0 ? x1 : x1 - 1;
    int x2 = x1 == imageWidth - 1 ? x1 : x1 + 1;
    int y0 = y1 == 0 ? y1 : y1 - 1;
    int y2 = y1 == imageHeight -1 ? y1 : y1 + 1;
 
    data[i1] = magnitude[i1];
    for (int x = x0; x <= x2; x++) {
      for (int y = y0; y <= y2; y++) {
        int i2 = x + y * imageWidth;
        if ((y != y1 || x != x1) && data[i2] == 0 && magnitude[i2] >= threshold) {
          follow(x, y, i2, threshold, depth + 1);
          return;
        }
      }
    }
  }

void computeGradients(float kernelRadius, int kernelWidth) {
    //generate the gaussian convolution masks
    float kernel[] = new float[kernelWidth];
    float diffKernel[] = new float[kernelWidth];
    int kwidth;
    for (kwidth = 0; kwidth < kernelWidth; kwidth++) {
      float g1 = gaussian(kwidth, kernelRadius);
      if (g1 <= GAUSSIAN_CUT_OFF && kwidth >= 2) break;
      float g2 = gaussian(kwidth - 0.5f, kernelRadius);
      float g3 = gaussian(kwidth + 0.5f, kernelRadius);
      kernel[kwidth] = (g1 + g2 + g3) / 3f / (2f * (float) Math.PI * kernelRadius * kernelRadius);
      diffKernel[kwidth] = g3 - g2;
    }

    int initX = kwidth - 1;
    int maxX = imageWidth - (kwidth - 1);
    int initY = imageWidth * (kwidth - 1);
    int maxY = imageWidth * (imageHeight - (kwidth - 1));
    
    //perform convolution in x and y directions
    for (int x = initX; x < maxX; x++) {
      for (int y = initY; y < maxY; y += imageWidth) {
        int index = x + y;
        float sumX = data[index] * kernel[0];
        float sumY = sumX;
        int xOffset = 1;
        int yOffset = imageWidth;
        for(; xOffset < kwidth ;) {
          sumY += kernel[xOffset] * (data[index - yOffset] + data[index + yOffset]);
          sumX += kernel[xOffset] * (data[index - xOffset] + data[index + xOffset]);
          yOffset += imageWidth;
          xOffset++;
        }
        
        yConv[index] = sumY;
        xConv[index] = sumX;
      }
 
    }
 
    for (int x = initX; x < maxX; x++) {
      for (int y = initY; y < maxY; y += imageWidth) {
        float sum = 0f;
        int index = x + y;
        for (int i = 1; i < kwidth; i++)
          sum += diffKernel[i] * (yConv[index - i] - yConv[index + i]);
 
        xGradient[index] = sum;
      }
 
    }

    for (int x = kwidth; x < imageWidth - kwidth; x++) {
      for (int y = initY; y < maxY; y += imageWidth) {
        float sum = 0.0f;
        int index = x + y;
        int yOffset = imageWidth;
        for (int i = 1; i < kwidth; i++) {
          sum += diffKernel[i] * (xConv[index - yOffset] - xConv[index + yOffset]);
          yOffset += imageWidth;
        }
 
        yGradient[index] = sum;
      }
 
    }
 
    initX = kwidth;
    maxX = imageWidth - kwidth;
    initY = imageWidth * kwidth;
    maxY = imageWidth * (imageHeight - kwidth);
    for (int x = initX; x < maxX; x++) {
      for (int y = initY; y < maxY; y += imageWidth) {
        int index = x + y;
        int indexN = index - imageWidth;
        int indexS = index + imageWidth;
        int indexW = index - 1;
        int indexE = index + 1;
        int indexNW = indexN - 1;
        int indexNE = indexN + 1;
        int indexSW = indexS - 1;
        int indexSE = indexS + 1;
        
        float xGrad = xGradient[index];
        float yGrad = yGradient[index];
        float gradMag = hypot(xGrad, yGrad);

        //perform non-maximal supression
        float nMag = hypot(xGradient[indexN], yGradient[indexN]);
        float sMag = hypot(xGradient[indexS], yGradient[indexS]);
        float wMag = hypot(xGradient[indexW], yGradient[indexW]);
        float eMag = hypot(xGradient[indexE], yGradient[indexE]);
        float neMag = hypot(xGradient[indexNE], yGradient[indexNE]);
        float seMag = hypot(xGradient[indexSE], yGradient[indexSE]);
        float swMag = hypot(xGradient[indexSW], yGradient[indexSW]);
        float nwMag = hypot(xGradient[indexNW], yGradient[indexNW]);
        float tmp;
        
        // An explanation of what's happening here, for those who want
        // to understand the source: This performs the "non-maximal
        // supression" phase of the Canny edge detection in which we
        // need to compare the gradient magnitude to that in the
        // direction of the gradient; only if the value is a local
        // maximum do we consider the point as an edge candidate.
        // 
        // We need to break the comparison into a number of different
        // cases depending on the gradient direction so that the
        // appropriate values can be used. To avoid computing the
        // gradient direction, we use two simple comparisons: first we
        // check that the partial derivatives have the same sign (1)
        // and then we check which is larger (2). As a consequence, we
        // have reduced the problem to one of four identical cases that
        // each test the central gradient magnitude against the values at
        // two points with 'identical support'; what this means is that
        // the geometry required to accurately interpolate the magnitude
        // of gradient function at those points has an identical
        // geometry (upto right-angled-rotation/reflection).
        // 
        // When comparing the central gradient to the two interpolated
        // values, we avoid performing any divisions by multiplying both
        // sides of each inequality by the greater of the two partial
        // derivatives. The common comparand is stored in a temporary
        // variable (3) and reused in the mirror case (4).

        if (xGrad * yGrad <= (float) 0 //(1)
          ? Math.abs(xGrad) >= Math.abs(yGrad) //(2)
            ? (tmp = Math.abs(xGrad * gradMag)) >= Math.abs(yGrad * neMag - (xGrad + yGrad) * eMag) //(3)
              && tmp > Math.abs(yGrad * swMag - (xGrad + yGrad) * wMag) //(4)
            : (tmp = Math.abs(yGrad * gradMag)) >= Math.abs(xGrad * neMag - (yGrad + xGrad) * nMag) //(3)
              && tmp > Math.abs(xGrad * swMag - (yGrad + xGrad) * sMag) //(4)
          : Math.abs(xGrad) >= Math.abs(yGrad) //(2)
            ? (tmp = Math.abs(xGrad * gradMag)) >= Math.abs(yGrad * seMag + (xGrad - yGrad) * eMag) //(3)
              && tmp > Math.abs(yGrad * nwMag + (xGrad - yGrad) * wMag) //(4)
            : (tmp = Math.abs(yGrad * gradMag)) >= Math.abs(xGrad * seMag + (yGrad - xGrad) * sMag) //(3)
              && tmp > Math.abs(xGrad * nwMag + (yGrad - xGrad) * nMag) //(4)
          ) {
          magnitude[index] = gradMag >= MAGNITUDE_LIMIT ? MAGNITUDE_MAX : (int) (MAGNITUDE_SCALE * gradMag);
          //NOTE: The orientation of the edge is not employed by this
          //implementation. It is a simple matter to compute it at
          //this point as: Math.atan2(yGrad, xGrad);
        } else {
          magnitude[index] = 0;
        }
      }
    }
  }
 

void doCannyEdgeDetection() {
    int i = 0;
    color c;

    // -------- Setup -------------
    updateProgress("Canny start");
    imageWidth = img.width;
    imageHeight = img.height;
    picsize = imageWidth * imageHeight;
    
    // Initialize arrays.
    if (data == null || picsize != data.length) {
      data = new int[picsize];
      magnitude = new int[picsize];

      xConv = new float[picsize];
      yConv = new float[picsize];
      xGradient = new float[picsize];
      yGradient = new float[picsize];
    }
    
    // -------- Read luminance -------------
    updateProgress("Canny readLuminance");
    i = 0;
    for( int y = 0; y < imageHeight; y ++ ) {
      for( int x = 0; x < imageWidth; x ++ ) {
        c = img.pixels[i]; // x + y*imageWidth
        // Average of      RED               GREEN           BLUE
        data[i++] = ((c >> 16 & 0xFF) + (c >> 8 & 0xFF) + (c & 0xFF)) / 3;
        // Should we use this?? data[i++] = Math.round(0.299f * r + 0.587f * g + 0.114f * b);
      }
    }

    // -------- Compute gradients -------------
    updateProgress("Canny compute gradients");
    computeGradients(gaussianKernelRadius, gaussianKernelWidth);
    
    // -------- Perform hysteresis -------------
    updateProgress("Canny hysteresis");
    int low = Math.round(lowThreshold * MAGNITUDE_SCALE);
    int high = Math.round( highThreshold * MAGNITUDE_SCALE);
    for (i = 0; i < picsize; i++)
      data[i] = 0;
    int offset = 0;
    for (int y = 0; y < imageHeight; y++) {
      for (int x = 0; x < imageWidth; x++) {
        if (data[offset] == 0 && magnitude[offset] >= high) {
          follow(x, y, offset, low, 0);
        }
        offset++;
      }
    }

    // -------- Threshold -------------
    updateProgress("Canny threshold");
    for (i = 0; i < picsize; i++) {
 //     data[i] = data[i] > 0 ? -1 : 0xff000000;
        data[i] = data[i] > 0 ? 0 : 0xFF;
    }

    // -------- Write out -------------
    updateProgress("Canny write");
    i = 0;
    for( int y = 0; y < imageHeight; y ++ )
      for( int x = 0; x < imageWidth; x ++ )
      {
        int d = data[i];
        edgeImg.pixels[i++] = color(d, d, d);
      }
    
    // -------- DONE -------------
    updateProgress("Canny done");


}
void drawImage() {
  
  int xx = (width - (324 * 2)) / 2;
  int yy = 150;
  
  lowThreshold = scrollBar[SBAR_LT].getPos();
  highThreshold = scrollBar[SBAR_HT].getPos();
  gaussianKernelRadius = scrollBar[SBAR_KR].getPos();
  gaussianKernelWidth = round(scrollBar[SBAR_KW].getPos());

  
  strokeWeight(4);
  stroke(#FF0000);
  noFill();
  rect(xx-2, yy-2, 324, 244);
  rect(xx+322, yy-2, 324, 244);
  strokeWeight(1);
  if(webcam.available()) {
    webcam.read();
  }
  image(webcam, xx, yy);
  img = webcam;
   img.loadPixels();
   text(img.width + "X" + img.height, xx, yy - 20);
  // Create an opaque image of the same size as the original
  if (edgeImg == null)
    edgeImg = createImage(img.width, img.height, RGB);
  
  if (!processingImage)
  {
    processingImage = true;
    thread("processImage");
  }
  
  image(edgeImg, xx + 324, yy); // Draw the new image
  
 
}

void processImage() {
  doCannyEdgeDetection();
  edgeImg.updatePixels();
  processingImage = false;
}

*/
